import os
import tempfile
from typing import Dict, Optional, Tuple

import numpy as np
from pydicom.dataset import FileDataset as DCMImage
import stir

from mu_map.file.interfile import (
    Interfile,
    InterfileKeys,
    load_interfile,
    write_interfile,
)


"""
A template for a header of a projection in INTERFILE format.
"""
TEMPLATE_HEADER_PROJ = """
INTERFILE  :=
imaging modality := nucmed
version of keys := 3.3
name of data file := {DATA_FILE}

GENERAL IMAGE DATA :=
type of data := Tomographic
imagedata byte order := LITTLEENDIAN
number format := float
number of bytes per pixel := 4
patient orientation := {PATIENT_ORIENTATION}
patient rotation := {PATIENT_ROTATION}

SPECT STUDY (General) := 
;matrix axis label [2] := axial coordinate
matrix size [2] := {SLICES}
scaling factor (mm/pixel) [2] := {SPACING_SLICES}
;matrix axis label [1] := bin coordinate
matrix size [1] := {BINS}
scaling factor (mm/pixel) [1] := {SPACING_BINS}
number of projections := {N_PROJECTIONS}
extent of rotation := {ROTATION}
process status := acquired

SPECT STUDY (acquired data) :=
direction of rotation := CW
start angle := {START_ANGLE}
orbit := circular
radius := {RADIUS}

END OF INTERFILE :=
"""


def forward_project(
    recon: Interfile,
    n_slices: Optional[int] = None,
    n_bins: Optional[int] = None,
    n_projections: int = 120,
    rotation: int = 360,
    start_angle: int = 180,
    radius: int = 150,
) -> Interfile:
    """
    Forward project a reconstruction in INTERFILE format.

    :param recon: the reconstruction in Interfile format
    :param n_slices: the number of slices of the projection - defaults to the number of slices of the reconstruction
    :param n_bins: the number of bins of the projection - defaults to the number of rows in the reconstruction
    :param n_projections: the number of views used for the projection
    :param rotation: the total rotation in degrees
    :param start_angle: the start angle of the rotation
    :param radius: the radius of the detector circle in mm
    :return: an interfile image as a tuple of a header and a numpy array
    """
    recon_header, recon_img = recon

    n_slices = recon_img.shape[0] if n_slices is None else n_slices
    n_bins = recon_img.shape[1] if n_bins is None else n_bins
    # work around a bug in STIR with start_angle = 0
    start_angle = 360 if start_angle == 0 else start_angle

    # prepare writing temporary files to configure the projection with STIR
    filename_recon = "recon.hv"
    filename_proj = "projection"
    filename_proj_data = f"{filename_proj}.s"
    filename_proj_header = f"{filename_proj}.hs"

    header_proj = TEMPLATE_HEADER_PROJ.replace("{DATA_FILE}", filename_proj_data)
    header_proj = header_proj.replace("{SLICES}", str(n_slices))
    header_proj = header_proj.replace("{BINS}", str(n_bins))
    header_proj = header_proj.replace(
        "{SPACING_SLICES}", recon_header[InterfileKeys.spacing(3)]
    )
    header_proj = header_proj.replace(
        "{SPACING_BINS}", recon_header[InterfileKeys.spacing(1)]
    )
    header_proj = header_proj.replace("{N_PROJECTIONS}", str(n_projections))
    header_proj = header_proj.replace("{ROTATION}", str(rotation))
    header_proj = header_proj.replace("{START_ANGLE}", str(start_angle))
    header_proj = header_proj.replace("{RADIUS}", str(radius))
    header_proj = header_proj.replace(
        "{PATIENT_ORIENTATION}", recon_header[InterfileKeys.patient_orientation]
    )
    header_proj = header_proj.replace(
        "{PATIENT_ROTATION}", recon_header[InterfileKeys.patient_rotation]
    )
    header_proj = header_proj.strip()

    # write temporary files to configure the projection with STIR
    dir_tmp = tempfile.TemporaryDirectory()

    filename_proj_header = os.path.join(dir_tmp.name, filename_proj_header)
    filename_recon = os.path.join(dir_tmp.name, filename_recon)

    with open(filename_proj_header, mode="w") as f:
        f.write(header_proj)
    with open(os.path.join(dir_tmp.name, filename_proj_data), mode="wb") as f:
        f.write(b"")
    write_interfile(filename_recon, recon)

    projdata = stir.ProjData.read_from_file(filename_proj_header)
    recon = stir.FloatVoxelsOnCartesianGrid.read_from_file(filename_recon)

    projmatrix = stir.ProjMatrixByBinUsingRayTracing()
    projmatrix.set_up(projdata.get_proj_data_info(), recon)

    forwardprojector = stir.ForwardProjectorByBinUsingProjMatrixByBin(projmatrix)

    projdataout = stir.ProjDataInMemory(
        projdata.get_exam_info(), projdata.get_proj_data_info()
    )

    forwardprojector.set_up(projdataout.get_proj_data_info(), recon)
    forwardprojector.forward_project(projdataout, recon)

    filename_out = os.path.join(dir_tmp.name, "out.hs")
    projdataout.write_to_file(filename_out)
    return load_interfile(filename_out)


if __name__ == "__main__":
    import argparse

    from mu_map.file.util import load_as_interfile

    parser = argparse.ArgumentParser(
        description="Forward project a reconstruction with STIR",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--recon",
        type=str,
        required=True,
        help="the reconstruction to project in INTERFILE or DICOM format",
    )
    parser.add_argument(
        "--out", type=str, required=True, help="the projection output by this script"
    )
    parser.add_argument(
        "--n_slices",
        type=int,
        help="configure a different number of slices than in the reconstruction",
    )
    parser.add_argument(
        "--n_projections",
        type=int,
        default=120,
        help="the number of performed projections",
    )
    parser.add_argument(
        "--rotation", type=int, default=360, help="the amount of rotation in degrees"
    )
    parser.add_argument(
        "--start_angle",
        type=int,
        default=180,
        help="the start angle for the projections",
    )
    parser.add_argument(
        "--radius", type=int, default=150, help="the radius of the rotation"
    )
    parser.add_argument(
        "-v", "--verbosity", type=int, default=0, help="configure the verbosity of STIR"
    )
    args = parser.parse_args()

    stir.Verbosity_set(args.verbosity)

    recon = load_as_interfile(args.recon)
    projection = forward_project(
        recon,
        n_slices=args.n_slices,
        n_projections=args.n_projections,
        rotation=args.rotation,
        start_angle=args.start_angle,
        radius=args.radius,
    )
    write_interfile(args.out, projection)
